searchNodes=[{"doc":"Utilities to work with Elixir source code. NOTICE: This library is under heavy development. Expect frequent breaking changes until the first stable v1.0 release is out. Installation Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.8&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected. Goals of the library Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No runtime dependencies, to simplify integration with other tools. Background There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. This discussion in the Elixir mailing list highlights the various issues faced when deciding if and how the comments would be preserved. Arjan Scherpenisse also did a talk where he discusses about the problems of using the standard Elixir AST to build refactoring tools. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest. Sourceror's AST Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropriate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , column : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , column : 4 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing comments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted.","ref":"Sourceror.html","title":"Sourceror","type":"module"},{"doc":"Appends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#append_comments/3","title":"Sourceror.append_comments/3","type":"function"},{"doc":"Compares two positions. Returns :gt if the first position comes after the second one, and :lt for vice versa. If the two positions are equal, :eq is returned. nil values for lines or columns are coalesced to 0 for integer comparisons.","ref":"Sourceror.html#compare_positions/2","title":"Sourceror.compare_positions/2","type":"function"},{"doc":"Shifts the line numbers of the node or metadata by the given line_correction . This function will update the :line , :closing , :do , :end and :end_of_expression line numbers of the node metadata if such fields are present.","ref":"Sourceror.html#correct_lines/3","title":"Sourceror.correct_lines/3","type":"function"},{"doc":"Returns the arguments of the node. iex&gt; Sourceror . get_args ( { :foo , [ ] , [ { :__block__ , [ ] , [ :ok ] } ] } ) [ { :__block__ , [ ] , [ :ok ] } ]","ref":"Sourceror.html#get_args/1","title":"Sourceror.get_args/1","type":"function"},{"doc":"Returns the column of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the column number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_column ( { :foo , [ column : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_column ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_column/2","title":"Sourceror.get_column/2","type":"function"},{"doc":"Returns the line where the given node ends. It recursively checks for end , closing and end_of_expression line numbers. If none is found, the default value is returned(defaults to 1). iex&gt; Sourceror . get_end_line ( { :foo , [ end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] ] , [ ] } ) 2 iex&gt; Sourceror . get_end_line ( { :foo , [ end_of_expression : [ line : 5 ] ] , [ ] } ) 5 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] , end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; &quot;&quot;&quot; ...&gt; alias Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) |&gt; Sourceror . get_end_line ( ) 3","ref":"Sourceror.html#get_end_line/2","title":"Sourceror.get_end_line/2","type":"function"},{"doc":"Returns the end position of the quoted expression. It recursively checks for end , closing and end_of_expression positions. If none is found, the default value is returned(defaults to [line: 1, column: 1] ). iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; A.{ ...&gt; B ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo( ...&gt; :a, ...&gt; :b ...&gt; ) ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 4 , column : 4 ]","ref":"Sourceror.html#get_end_position/2","title":"Sourceror.get_end_position/2","type":"function"},{"doc":"Returns the line of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the line number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_line ( { :foo , [ line : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_line ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_line/2","title":"Sourceror.get_line/2","type":"function"},{"doc":"Returns the metadata of the given node. iex&gt; Sourceror . get_meta ( { :foo , [ line : 5 ] , [ ] } ) [ line : 5 ]","ref":"Sourceror.html#get_meta/1","title":"Sourceror.get_meta/1","type":"function"},{"doc":"Gets the range used by the given quoted expression in the source code. The quoted expression must have at least line and column metadata, otherwise it is not possible to calculate an accurate range, or to calculate it at all. This function is most useful when used after Sourceror.parse_string/1 , before any kind of modification to the AST. The range is a map with :start and :end positions. iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; def foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 4 ] } iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 2 ] }","ref":"Sourceror.html#get_range/1","title":"Sourceror.get_range/1","type":"function"},{"doc":"Returns the start position of a node. iex&gt; quoted = Sourceror . parse_string! ( &quot; :foo&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 2 ] iex&gt; quoted = Sourceror . parse_string! ( &quot; \\n \\n foo()&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;Foo.{Bar}&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo[:bar]&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo(:bar)&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ]","ref":"Sourceror.html#get_start_position/2","title":"Sourceror.get_start_position/2","type":"function"},{"doc":"Parses a single expression from the given string. It tries to parse on a per-line basis. Returns {:ok, quoted, rest} on success or {:error, source} on error. Examples iex&gt; ~S&quot;&quot;&quot; ...&gt; 42 ...&gt; ...&gt; :ok ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_expression ( ) { :ok , { :__block__ , [ trailing_comments : [ ] , leading_comments : [ ] , token : &quot;42&quot; , line : 2 , column : 1 ] , [ 42 ] } , &quot; \\n :ok&quot; } Options :from_line - The line at where the parsing should start. Defaults to 1 .","ref":"Sourceror.html#parse_expression/2","title":"Sourceror.parse_expression/2","type":"function"},{"doc":"Parses the source code into an extended AST suitable for source manipulation as described in Code.quoted_to_algebra/2 . Two additional fields are added to nodes metadata: :leading_comments - a list holding the comments found before the node. :trailing_comments - a list holding the comments found before the end of the node. For example, comments right before the end keyword. Comments are the same maps returned by Code.string_to_quoted_with_comments/2 .","ref":"Sourceror.html#parse_string/1","title":"Sourceror.parse_string/1","type":"function"},{"doc":"Same as parse_string/1 but raises on error.","ref":"Sourceror.html#parse_string!/1","title":"Sourceror.parse_string!/1","type":"function"},{"doc":"Applies one or more patches to the given string. This functions limits itself to apply the patches in order, but it does not check for overlapping ranges, so make sure to pass non-overlapping patches. A patch is a map containing at least the range that it should patch, and the change to be applied in the range, for example: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; if not allowed? do ...&gt; raise &quot;Not allowed!&quot; ...&gt; end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;unless allowed? do \\n raise \\&quot; Not allowed! \\&quot; \\n end&quot; , ...&gt; range : %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 4 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; unless allowed? do raise &quot;Not allowed!&quot; end &quot;&quot;&quot; A range can also be a function, in which case the original text in the patch range will be given as an argument: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; hello :world ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &amp; String . upcase / 1 , ...&gt; range : %{ start : [ line : 1 , column : 7 ] , end : [ line : 1 , column : 13 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; hello :WORLD &quot;&quot;&quot; By default, the patch will be automatically indented to match the indentation of the range it wants to replace if the change is a text string: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; foo do bar do ...&gt; :ok ...&gt; end end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;baz do \\n :not_ok \\n end&quot; , ...&gt; range : %{ start : [ line : 1 , column : 8 ] , end : [ line : 3 , column : 6 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; foo do baz do :not_ok end end &quot;&quot;&quot; If you don't want this behavior, you can add :preserve_indentation: false to your patch: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; foo do bar do ...&gt; :ok ...&gt; end end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;baz do \\n :not_ok \\n end&quot; , ...&gt; range : %{ start : [ line : 1 , column : 8 ] , end : [ line : 3 , column : 6 ] } , ...&gt; preserve_indentation : false ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; foo do baz do :not_ok end end &quot;&quot;&quot;","ref":"Sourceror.html#patch_string/2","title":"Sourceror.patch_string/2","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression. See postwalk/3 for more information.","ref":"Sourceror.html#postwalk/2","title":"Sourceror.postwalk/2","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.postwalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#postwalk/3","title":"Sourceror.postwalk/3","type":"function"},{"doc":"Prepends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#prepend_comments/3","title":"Sourceror.prepend_comments/3","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression. See prewalk/3 for more information.","ref":"Sourceror.html#prewalk/2","title":"Sourceror.prewalk/2","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.prewalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#prewalk/3","title":"Sourceror.prewalk/3","type":"function"},{"doc":"A wrapper around Code.quoted_to_algebra/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#quoted_to_algebra/2","title":"Sourceror.quoted_to_algebra/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted/2","title":"Sourceror.string_to_quoted/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments!/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted!/2","title":"Sourceror.string_to_quoted!/2","type":"macro"},{"doc":"Converts a quoted expression to a string. The comments line number will be ignored and the line number of the associated node will be used when formatting the code. Options :line_length - The max line length for the formatted code. :indent - how many indentations to insert at the start of each line. Note that this only prepends the indents without checking the indentation of nested blocks. Defaults to 0 . :indent_type - the type of indentation to use. It can be one of :spaces , :single_space or :tabs . Defaults to :spaces . :format - if set to :splicing , if the quoted expression is a list, it will strip the square brackets. This is useful to print a single element of a keyword list.","ref":"Sourceror.html#to_string/2","title":"Sourceror.to_string/2","type":"function"},{"doc":"Updates the arguments for the given node. iex&gt; node = { :foo , [ line : 1 ] , [ { :__block__ , [ line : 1 ] , [ 2 ] } ] } iex&gt; updater = fn args -&gt; Enum . map ( args , &amp; Sourceror . correct_lines ( &amp;1 , 2 ) ) end iex&gt; Sourceror . update_args ( node , updater ) { :foo , [ line : 1 ] , [ { :__block__ , [ line : 3 ] , [ 2 ] } ] }","ref":"Sourceror.html#update_args/2","title":"Sourceror.update_args/2","type":"function"},{"doc":"","ref":"Sourceror.html#t:comment/0","title":"Sourceror.comment/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:patch/0","title":"Sourceror.patch/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:position/0","title":"Sourceror.position/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:range/0","title":"Sourceror.range/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:traversal_function/0","title":"Sourceror.traversal_function/0","type":"type"},{"doc":"Utilities to merge an un-merge comments and quoted expressions.","ref":"Sourceror.Comments.html","title":"Sourceror.Comments","type":"module"},{"doc":"Does the opposite of merge_comments/2 , it extracts the comments from the quoted expression and returns both as a {quoted, comments} tuple.","ref":"Sourceror.Comments.html#extract_comments/2","title":"Sourceror.Comments.extract_comments/2","type":"function"},{"doc":"Merges the comments into the given quoted expression. The comments are inserted into the metadata of their closest node. Comments in the same line of before a node are inserted into the :leading_comments field while comments that are right before an end keyword are inserted into the :trailing_comments field.","ref":"Sourceror.Comments.html#merge_comments/2","title":"Sourceror.Comments.merge_comments/2","type":"function"},{"doc":"The state struct for Sourceror traversal functions.","ref":"Sourceror.TraversalState.html","title":"Sourceror.TraversalState","type":"module"},{"doc":"","ref":"Sourceror.TraversalState.html#t:t/0","title":"Sourceror.TraversalState.t/0","type":"type"},{"doc":"Implements a Zipper for the Elixir AST based on Gérard Huet Functional pearl: the zipper paper and Clojure's clojure.zip API. A zipper is a data structure that represents a location in a tree from the perspective of the current node, also called focus . It is represented by a 2-tuple where the first element is the focus and the second element is the metadata/context. When the focus is the topmost node, the metadata is nil , or :end after the end of a traversal.","ref":"Sourceror.Zipper.html","title":"Sourceror.Zipper","type":"module"},{"doc":"Inserts the item as the rightmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#append_child/2","title":"Sourceror.Zipper.append_child/2","type":"function"},{"doc":"Returns true if the node is a branch.","ref":"Sourceror.Zipper.html#branch?/1","title":"Sourceror.Zipper.branch?/1","type":"function"},{"doc":"Returns a list of children of the node.","ref":"Sourceror.Zipper.html#children/1","title":"Sourceror.Zipper.children/1","type":"function"},{"doc":"Returns the zipper of the leftmost child of the node at this zipper, or nil if no there's no children.","ref":"Sourceror.Zipper.html#down/1","title":"Sourceror.Zipper.down/1","type":"function"},{"doc":"Returns true if the zipper represents the end of a depth-first walk.","ref":"Sourceror.Zipper.html#end?/1","title":"Sourceror.Zipper.end?/1","type":"function"},{"doc":"Returns a zipper to the node that satisfies the predicate function, or nil if none is found.","ref":"Sourceror.Zipper.html#find/2","title":"Sourceror.Zipper.find/2","type":"function"},{"doc":"Inserts the item as the leftmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#insert_child/2","title":"Sourceror.Zipper.insert_child/2","type":"function"},{"doc":"Inserts the item as the left sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_left/2","title":"Sourceror.Zipper.insert_left/2","type":"function"},{"doc":"Inserts the item as the right sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_right/2","title":"Sourceror.Zipper.insert_right/2","type":"function"},{"doc":"Returns the zipper of the left sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#left/1","title":"Sourceror.Zipper.left/1","type":"function"},{"doc":"Returns the leftmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#leftmost/1","title":"Sourceror.Zipper.leftmost/1","type":"function"},{"doc":"Returns a new branch node, given an existing node and new children.","ref":"Sourceror.Zipper.html#make_node/2","title":"Sourceror.Zipper.make_node/2","type":"function"},{"doc":"Returns the following zipper in depth-first pre-order. When reaching the end, returns a distinguished zipper detectable via end?/1 . If it's already at the end, it stays there.","ref":"Sourceror.Zipper.html#next/1","title":"Sourceror.Zipper.next/1","type":"function"},{"doc":"Returns the node at the zipper.","ref":"Sourceror.Zipper.html#node/1","title":"Sourceror.Zipper.node/1","type":"function"},{"doc":"Returns the previous zipper in depth-first pre-order. If it's already at the end, it returns nil.","ref":"Sourceror.Zipper.html#prev/1","title":"Sourceror.Zipper.prev/1","type":"function"},{"doc":"Removes the node at the zipper, returning the zipper that would have preceded it in a depth-first walk.","ref":"Sourceror.Zipper.html#remove/1","title":"Sourceror.Zipper.remove/1","type":"function"},{"doc":"Replaces the current node in the zipper with a new node.","ref":"Sourceror.Zipper.html#replace/2","title":"Sourceror.Zipper.replace/2","type":"function"},{"doc":"&quot;Returns the zipper of the right sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#right/1","title":"Sourceror.Zipper.right/1","type":"function"},{"doc":"Returns the rightmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#rightmost/1","title":"Sourceror.Zipper.rightmost/1","type":"function"},{"doc":"Walks the zipper all the way up and returns the root node.","ref":"Sourceror.Zipper.html#root/1","title":"Sourceror.Zipper.root/1","type":"function"},{"doc":"Walks the zipper all the way up and returns the top zipper.","ref":"Sourceror.Zipper.html#top/1","title":"Sourceror.Zipper.top/1","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node. Returns a zipper to the root node. The function must return a zipper.","ref":"Sourceror.Zipper.html#traverse/2","title":"Sourceror.Zipper.traverse/2","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node with an accumulator. Returns a zipper to the root node and the accumulator.","ref":"Sourceror.Zipper.html#traverse/3","title":"Sourceror.Zipper.traverse/3","type":"function"},{"doc":"Returns the zipper of the parent of the node at this zipper, or nil if at the top.","ref":"Sourceror.Zipper.html#up/1","title":"Sourceror.Zipper.up/1","type":"function"},{"doc":"Replaces the current node in the zipper with the result of applying fun to the node.","ref":"Sourceror.Zipper.html#update/2","title":"Sourceror.Zipper.update/2","type":"function"},{"doc":"Creates a zipper from a tree node.","ref":"Sourceror.Zipper.html#zip/1","title":"Sourceror.Zipper.zip/1","type":"function"},{"doc":"","ref":"Sourceror.Zipper.html#t:path/0","title":"Sourceror.Zipper.path/0","type":"type"},{"doc":"","ref":"Sourceror.Zipper.html#t:tree/0","title":"Sourceror.Zipper.tree/0","type":"type"},{"doc":"","ref":"Sourceror.Zipper.html#t:zipper/0","title":"Sourceror.Zipper.zipper/0","type":"type"},{"doc":"Changelog for Sourceror v0.8 All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Fixed issues that caused comments to be misplaced. [Sourceror] Updated internal normalizer to match latest Elixir 1.13 version.","ref":"changelog.html#v0-8-8-2021-11-27","title":"Changelog - v0.8.8 (2021-11-27)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Fixed an issue that caused newlines to be wrongly removed.","ref":"changelog.html#v0-8-7-2021-10-23","title":"Changelog - v0.8.7 (2021-10-23)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Fixed an issue that caused comments in pipelines to be misplaced.","ref":"changelog.html#v0-8-6-2021-10-23","title":"Changelog - v0.8.6 (2021-10-23)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Fixed issue that prevented keyword lists from preserving their original format in tuples.","ref":"changelog.html#v0-8-5-2021-10-1","title":"Changelog - v0.8.5 (2021-10-1)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] get_range/1 now properly handles naked AST lists, like the ones coming from partial keyword lists, or stabs like a -&gt; b .","ref":"changelog.html#v0-8-4-2021-09-15","title":"Changelog - v0.8.4 (2021-09-15)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] get_range/1 now handles partial keyword list syntax instead of crashing.","ref":"changelog.html#v0-8-3-2021-09-13","title":"Changelog - v0.8.3 (2021-09-13)","type":"extras"},{"doc":"1. Bug fixes [Sourceror.Zipper] down/1 now correctly uses nil as the right siblings if the branch node has a single child.","ref":"changelog.html#v0-8-2-2021-08-12","title":"Changelog - v0.8.2 (2021-08-12)","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Sourceror.get_range/1 now correctly calculates the range when there is a comment in the same line as the node.","ref":"changelog.html#v0-8-1-2021-07-03","title":"Changelog - v0.8.1 (2021-07-03)","type":"extras"},{"doc":"1. Enhancements [Sourceror] Added Sourceror.patch_string/2 [Sourceror] Added the format: :splicing option to Sourceror.to_string/2 2. Bug fixes [Sourceror] Now Sourceror.to_string/2 won't produce invalid Elixir code when a keyword list element is at the beginning of a non-keyword list. [Sourceror] Now Sourceror.get_range/1 will take the leading comments into account when calculating the range.","ref":"changelog.html#v0-8-0-2021-06-24","title":"Changelog - v0.8.0 (2021-06-24)","type":"extras"},{"doc":"The changelog for v0.7 releases can be found in the v0.7 branch .","ref":"changelog.html#v0-7-0-2021-06-12","title":"Changelog - v0.7.0 (2021-06-12)","type":"extras"},{"doc":"Contributing to Sourceror","ref":"contributing.html","title":"Contributing","type":"extras"},{"doc":"For issues like bug reports please file an issue in the Issue tracker .","ref":"contributing.html#issues","title":"Contributing - Issues","type":"extras"},{"doc":"To propose new features to the library, please open a thread in the Discussions section under the Proposals category. Development Contributions Take a look at the open issues here and post that you plan on working on an issue, so we can discuss approaches and make sure there isn't duplicated work happening. Before submitting a pull request, make sure all of the following pass: mix deps.get mix check","ref":"contributing.html#proposals","title":"Contributing - Proposals","type":"extras"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS","ref":"license.html","title":"License","type":"extras"},{"doc":"Sourceror 🧙 Utilities to work with Elixir source code. NOTICE: This library is under heavy development. Expect frequent breaking changes until the first stable v1.0 release is out.","ref":"readme.html","title":"Overview","type":"extras"},{"doc":"Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.8&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected.","ref":"readme.html#installation","title":"Overview - Installation","type":"extras"},{"doc":"Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No runtime dependencies, to simplify integration with other tools.","ref":"readme.html#goals-of-the-library","title":"Overview - Goals of the library","type":"extras"},{"doc":"There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. This discussion in the Elixir mailing list highlights the various issues faced when deciding if and how the comments would be preserved. Arjan Scherpenisse also did a talk where he discusses about the problems of using the standard Elixir AST to build refactoring tools. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest.","ref":"readme.html#background","title":"Overview - Background","type":"extras"},{"doc":"Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropriate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , column : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , column : 4 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing comments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted.","ref":"readme.html#sourceror-s-ast","title":"Overview - Sourceror's AST","type":"extras"},{"doc":"You can find Sourceror documentation on Hex Docs .","ref":"readme.html#documentation","title":"Overview - Documentation","type":"extras"},{"doc":"You can find usage examples in the examples folder. You can run them by cloning the repo and running elixir examples/&lt;example_file&gt;.exs . You can also find documented examples you can run with Livebook in the notebooks folder.","ref":"readme.html#examples","title":"Overview - Examples","type":"extras"},{"doc":"If you want to contribute to Sourceror, please check our Contributing section for pointers.","ref":"readme.html#contributing","title":"Overview - Contributing","type":"extras"},{"doc":"If you have any questions about Sourceror or need assistance, please open a thread in the Discussions section.","ref":"readme.html#getting-assistance","title":"Overview - Getting assistance","type":"extras"},{"doc":"Copyright (c) 2021 dorgandash@gmail.com Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"readme.html#copyright-and-license","title":"Overview - Copyright and License","type":"extras"}]